//======================================================================================//
// This software is based on Robert Walter's code (github.com/robbeofficial)            //
// Usage:   ./NiViewSKL CX_Y_Z.oni                                                      //
//          Note: create a folder named "out_hand_sequence" to store the jpeg sequence. //
//======================================================================================//

#include <iostream>
#include <sstream>
#include <fstream>
using namespace std;

#include <opencv/cv.h>
#include <opencv/highgui.h>
using namespace cv;

#include "functions.h"

#include <XnCppWrapper.h>
using namespace xn;

/*//////////////////////////////////////////////////////////////////////
// constants
//////////////////////////////////////////////////////////////////////*/

const Size frameSize(640, 480);
const unsigned int maxUsers = 20;
const Mat emptyMat();

const int maxHandRadius = 100; // in px (original : 128)
const short handDepthRange = 200; // in mm (original das capturas: 200)

/*//////////////////////////////////////////////////////////////////////
// globals
//////////////////////////////////////////////////////////////////////*/

//Recorder
xn::Recorder record;
ofstream skelFile;

//to read bones from file..
ifstream skelInfo;

// openNI context and generator nodes
Context context;
ImageGenerator imageGen;
DepthGenerator depthGen;
IRGenerator irGen;
UserGenerator userGen;

// frame buffers
Mat bgrMat(frameSize, CV_8UC3);
Mat depthMat(frameSize, CV_16UC1);
Mat depthMat8(frameSize, CV_8UC1);
Mat depthMatBgr(frameSize, CV_8UC3);

/*//////////////////////////////////////////////////////////////////////
// callbacks
//////////////////////////////////////////////////////////////////////*/

void XN_CALLBACK_TYPE onNewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	cout << "onNewUser(" << nId << ")" << endl;
	SkeletonCapability skelCap = userGen.GetSkeletonCap();

	// load default calibration data
	skelCap.LoadCalibrationDataFromFile(nId, "UserCalibration.bin");

	// instantly track skeleton
	skelCap.StartTracking(nId);
}

void XN_CALLBACK_TYPE onLostUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	printf("Lost user %d\n", nId);
}

/*//////////////////////////////////////////////////////////////////////
// functions
//////////////////////////////////////////////////////////////////////*/

const Scalar skeletonColor(255,255,255);

//this function is used to draw a bone on the screen
void drawBone(XnUserID player,XnSkeletonJoint eJoint1, XnSkeletonJoint eJoint2)
{
	XnSkeletonJointPosition joint1, joint2;
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint1, joint1);
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint2, joint2);

	if (joint1.fConfidence < 0.5 || joint2.fConfidence < 0.5)
	{
		return;
	}

	XnPoint3D pt[2];
	pt[0] = joint1.position;
	pt[1] = joint2.position;

	depthGen.ConvertRealWorldToProjective(2, pt, pt);

	line(depthMatBgr, Point(pt[0].X, pt[0].Y), Point(pt[1].X, pt[1].Y), skeletonColor);
}

//this function is used to write a bone to the .skl file
void writeBone(XnUserID player,XnSkeletonJoint eJoint)
{
    XnSkeletonJointPosition joint;
	userGen.GetSkeletonCap().GetSkeletonJointPosition(player, eJoint, joint);

	if (joint.fConfidence < 0.5)
	{
		skelFile << "1000 1000 ";
	}
	else
	{
	    XnPoint3D pt[1];
        pt[0] = joint.position;
        depthGen.ConvertRealWorldToProjective(1, pt, pt);
        skelFile << pt[0].X << " " << pt[0].Y << " ";
	}
}

//this function draws the skeleton from kinect data
void drawSkeleton(XnUserID player)
{
	drawBone(player, XN_SKEL_HEAD, XN_SKEL_NECK);

	drawBone(player, XN_SKEL_NECK, XN_SKEL_LEFT_SHOULDER);
	drawBone(player, XN_SKEL_LEFT_SHOULDER, XN_SKEL_LEFT_ELBOW);
	drawBone(player, XN_SKEL_LEFT_ELBOW, XN_SKEL_LEFT_HAND);

	drawBone(player, XN_SKEL_NECK, XN_SKEL_RIGHT_SHOULDER);
	drawBone(player, XN_SKEL_RIGHT_SHOULDER, XN_SKEL_RIGHT_ELBOW);
	drawBone(player, XN_SKEL_RIGHT_ELBOW, XN_SKEL_RIGHT_HAND);

	drawBone(player, XN_SKEL_LEFT_SHOULDER, XN_SKEL_TORSO);
	drawBone(player, XN_SKEL_RIGHT_SHOULDER, XN_SKEL_TORSO);
}

//simpler version of drawBone
void drawBone(XnPoint3D pt1, XnPoint3D pt2)
{
    line(depthMatBgr, Point(pt1.X, pt1.Y), Point(pt2.X, pt2.Y), skeletonColor);
}

int oldFrame = 0;
string oldLine = "";

float skTorso[3];
float skRHand[3];
float skLHand[3];

//this function draws the skeleton from SKL file
void drawSkeletonFromFile(int currentFrame)
{
    XnPoint3D pt[24];

    std::string line;

    if(oldFrame != currentFrame)
    {
        getline (skelInfo,line);
        oldLine = line;
        oldFrame = currentFrame;
    }
    else
        line = oldLine;


    stringstream ss(line);
    string s;

    getline(ss, s, ' ');//frame

    getline(ss, s, ' ');
    pt[XN_SKEL_HEAD].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_HEAD].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_NECK].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_NECK].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_TORSO].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_TORSO].Y = atoi(s.c_str());
    skTorso[0]=pt[XN_SKEL_TORSO].X; skTorso[1]=pt[XN_SKEL_TORSO].Y;

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_SHOULDER].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_SHOULDER].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_ELBOW].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_ELBOW].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_HAND].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_LEFT_HAND].Y = atoi(s.c_str());
    skLHand[0]=pt[XN_SKEL_LEFT_HAND].X; skLHand[1]=pt[XN_SKEL_LEFT_HAND].Y;

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_SHOULDER].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_SHOULDER].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_ELBOW].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_ELBOW].Y = atoi(s.c_str());

    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_HAND].X = atoi(s.c_str());
    getline(ss, s, ' ');
    pt[XN_SKEL_RIGHT_HAND].Y = atoi(s.c_str());
    skRHand[0]=pt[XN_SKEL_RIGHT_HAND].X; skRHand[1]=pt[XN_SKEL_RIGHT_HAND].Y;

    drawBone(pt[XN_SKEL_HEAD], pt[XN_SKEL_NECK]);

	drawBone(pt[XN_SKEL_NECK], pt[XN_SKEL_LEFT_SHOULDER]);
	drawBone(pt[XN_SKEL_LEFT_SHOULDER], pt[XN_SKEL_LEFT_ELBOW]);
	drawBone(pt[XN_SKEL_LEFT_ELBOW], pt[XN_SKEL_LEFT_HAND]);

	drawBone(pt[XN_SKEL_NECK], pt[XN_SKEL_RIGHT_SHOULDER]);
	drawBone(pt[XN_SKEL_RIGHT_SHOULDER], pt[XN_SKEL_RIGHT_ELBOW]);
	drawBone(pt[XN_SKEL_RIGHT_ELBOW], pt[XN_SKEL_RIGHT_HAND]);

	drawBone(pt[XN_SKEL_LEFT_SHOULDER], pt[XN_SKEL_TORSO]);
	drawBone(pt[XN_SKEL_RIGHT_SHOULDER], pt[XN_SKEL_TORSO]);
}

//this function writes the skeleton to the SKL file
void writeSkeleton(XnUserID player, int currentFrame)
{
    skelFile << convertInt(currentFrame) << " ";

    writeBone(player,  XN_SKEL_HEAD);

	writeBone(player, XN_SKEL_NECK);
	writeBone(player, XN_SKEL_TORSO);

	writeBone(player, XN_SKEL_LEFT_SHOULDER);
	writeBone(player, XN_SKEL_LEFT_ELBOW);
    writeBone(player, XN_SKEL_LEFT_HAND);

	writeBone(player, XN_SKEL_RIGHT_SHOULDER);
	writeBone(player, XN_SKEL_RIGHT_ELBOW);
	writeBone(player, XN_SKEL_RIGHT_HAND);

	skelFile << "\n";
}

//kinect default init function
void initKinect(bool initImage, bool initDepth, bool initIR, bool initUser, char* oniFile) {
	XnStatus nRetVal = XN_STATUS_OK;

	// Initialize context object
	nRetVal = context.Init();
	if(oniFile != "") context.OpenFileRecording(oniFile);
	cout << "init : " << xnGetStatusString(nRetVal) << endl;

	// default output mode
	XnMapOutputMode outputMode;
	outputMode.nXRes = 640;
	outputMode.nYRes = 480;
	outputMode.nFPS = 30;

	// Create an ImageGenerator node
	if (initImage) {
		nRetVal = imageGen.Create(context);
		cout << "imageGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = imageGen.SetMapOutputMode(outputMode);
		cout << "imageGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = imageGen.GetMirrorCap().SetMirror(true);
		cout << "imageGen.GetMirrorCap().SetMirror : " << xnGetStatusString(nRetVal) << endl;
	}

	// Create a DepthGenerator node
	if (initDepth) {
		nRetVal = depthGen.Create(context);
		cout << "depthGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = depthGen.SetMapOutputMode(outputMode);
		cout << "depthGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = depthGen.GetMirrorCap().SetMirror(true);
		cout << "depthGen.GetMirrorCap().SetMirror : " << xnGetStatusString(nRetVal) << endl;
	}

	// Create an IRGenerator node
	if (initIR) {
		nRetVal = irGen.Create(context);
		cout << "irGen.Create : " << xnGetStatusString(nRetVal) << endl;
		nRetVal = irGen.SetMapOutputMode(outputMode);
		cout << "irGen.SetMapOutputMode : " << xnGetStatusString(nRetVal) << endl;
	}

	// create user generator
	if (initUser) {
		nRetVal = userGen.Create(context);
		cout << "userGen.Create : " << xnGetStatusString(nRetVal) << endl;
		userGen.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);
	}

	// Make it start generating data
	nRetVal = context.StartGeneratingAll();
	cout << "context.StartGeneratingAll : " << xnGetStatusString(nRetVal) << endl;
}

//returns the joint confidence based on the skeleton
float getJointImgCoordinates(const SkeletonCapability &skeletonCapability, const XnUserID userId, const XnSkeletonJoint skeletonJoint, float *v) {
	XnVector3D projective;
	XnSkeletonJointPosition skeletonJointPosition;
	skeletonCapability.GetSkeletonJointPosition(userId, skeletonJoint, skeletonJointPosition);

	depthGen.ConvertRealWorldToProjective(1, &skeletonJointPosition.position, &projective);

	v[0] = projective.X;
	v[1] = projective.Y;
	v[2] = projective.Z / 1000.0f;

	return skeletonJointPosition.fConfidence;
}

//Save frame hand shape as jpeg image (used to convert .oni videos to a jpeg sequence)
string getFrameName(int number)
{
   stringstream ss;//create a stringstream
   ss << number;//add number to the stream

   string frameName = "out_hand_sequence/out_";
   frameName += ss.str();
   frameName += ".jpg";

   return frameName;
}
int frameCounter = 0;

//hand segmentation
bool getHandContour(const Mat &depthMat, const float *v, vector<Point> &handContour) {
	const double epsilon = 17.5; // approximation accuracy (maximum distance between the original hand contour and its approximation)

	//unsigned short depth = (unsigned short) (v[2] * 1000.0f); // hand depth
	unsigned short near = 0; // near clipping plane -> original : depth - 100
	unsigned short far = 1500; // far clipping plane -> original : depth + 100

	static Mat mask(frameSize, CV_8UC1);
	mask.setTo(0);

	// extract hand region
	circle(mask, Point(v[0], v[1]), maxHandRadius, 255, CV_FILLED);
	mask = mask & depthMat > near & depthMat < far;

	// DEBUG(show mask)
	imshow("mask1", mask);
	imwrite(getFrameName(frameCounter++),mask);

	// assume largest contour in hand region to be the hand contour
	vector<vector<Point> > contours;
	findContours(mask, contours, CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE);
	int n = contours.size();
	int maxI = -1;
	int maxSize = -1;
	for (int i=0; i<n; i++) {
		int size  = contours[i].size();
		if (size > maxSize) {
			maxSize = size;
			maxI = i;
		}
	}

	bool handContourFound = (maxI >= 0);

	if (handContourFound) {
		approxPolyDP( Mat(contours[maxI]), handContour, epsilon, true );
		//handContour = contours[maxI];
	}

	// DEBUG(draw hand contour
	mask.setTo(0);
	if (maxI >= 0) {
		vector<vector<Point> > contours2;
		contours2.push_back(contours[maxI]);
		drawContours(mask, contours2, -1, 255);
		//imshow("mask2", mask); //uncomment to show image debug
	}

	return maxI >= 0;
}

//draw hand contour (just for debug propose)
void drawContour(Mat &img, const vector<Point> &contour, const Scalar &color) {
	vector<vector<Point> > contours;
	contours.push_back(contour);
	drawContours(img, contours, -1, color);
}

//Return the depth factor from a given pixel
int getDepthAt(Mat mat, int x, int y)
{
    struct RGB { unsigned char b, g, r; };

    assert((mat.step/mat.cols) == sizeof(RGB));
    RGB *data = (RGB*)mat.data;
    data += y * mat.cols + x;

    //printf("COLOR: %d  %d  %d\n",data->r,data->g,data->b);
    return data->b;
}

//main program
int main(int argc, char **argv) {
	const float minHandExtension = 0.2f; // in meters
	const double grabConvexity = 0.8;

	XnUserID userIds[maxUsers] = {0};
	XnUInt16 nUsers = maxUsers;
	char key = 0;

    bool readBones = false;

	// init context and generators
	if(argc>1)
	{
	    cout << "Initializing:" << argv[1] << endl;
	    initKinect(true, true, false, true, argv[1]);

	    //create output folder: out_hand_sequence/
	    cout << "to store the result please create the folder: out_hand_sequence" << endl;

        //load bones file
	    string skelFilename = argv[1];
	    skelFilename = skelFilename.substr(0,skelFilename.size()-3);//remove the oni extension
	    skelFilename+="skl";//add skl
	    skelInfo.open(skelFilename.c_str());

        cout << "Read bones from:" << skelFilename << endl;

	    readBones = true;
	}
	else
	  initKinect(true, true, false, true, "");

	// register some handy callbacks
	XnCallbackHandle hUserCallbacks;
	userGen.RegisterUserCallbacks(onNewUser, onLostUser, NULL, hUserCallbacks);

	SkeletonCapability skelCap = userGen.GetSkeletonCap();

	Mat mask(frameSize, CV_8UC1);

	int oldUsers = 0;

	bool recording = false;
    int currentFrame=0;
    bool paused = false;

    if(!argc>1)
	{
        setupSamplesWindow();
        loadSigns();
        paused = true;
	}

	std::string userName;
	if(argc<=1)
	{
	    cout << "User:";
        cin >> userName;
	}


    //calibrate kinect
    //http://www.informatik.uni-freiburg.de/~engelhar/calibration.html
	CvMat * camera_matrix = (CvMat*)cvLoad("Intrinsics_RGB.xml");
    CvMat * distortion_coeffs = (CvMat*)cvLoad("Distortion_RGB.xml");

    Mat undistortedImg(frameSize, CV_8UC3);

	while ( (key = (char) waitKey(1)) != 27 ) {

		if(!paused) context.WaitAndUpdateAll();

		// acquire bgr image
		{
			Mat mat(frameSize, CV_8UC3, (unsigned char*) imageGen.GetImageMap());
			cvtColor(mat, bgrMat, CV_RGB2BGR);
		}

		// acquire depth image
		{
			Mat mat(frameSize, CV_16UC1, (unsigned char*) depthGen.GetDepthMap());
			mat.copyTo(depthMat);
			depthMat.convertTo(depthMat8, CV_8UC1, 255.0f / 3000.0f);
			cvtColor(depthMat8, depthMatBgr, CV_GRAY2BGR);
		}

		// iterate over all user
		nUsers = userGen.GetNumberOfUsers();
		userGen.GetUsers(userIds, nUsers);
		float conf; // joint confidence
		float rh[3]; // right hand coordinates (x[px], y[px], z[meters])
		float lh[3]; // left hand coordinates
		float t[3]; // torso coordinates

        if(readBones)
        {

            if(!paused) currentFrame++;

            drawSkeletonFromFile(currentFrame);

            if(key == 'p')//pause playback
            {
                paused = 1-paused;
            }

            if(key == 'n')
            {
                context.WaitAndUpdateAll();
                currentFrame++;
            }

            //draw torso circle
            unsigned char shade = 255 - (unsigned char)( getDepthAt(depthMatBgr,skTorso[0],skTorso[1]) *  128.0f);
            circle(depthMatBgr, Point(skTorso[0], skTorso[1]), 10, Scalar(shade, 0, 0), -1);

            //perform hand segmentation
            DepthMetaData depthMD;
            depthGen.GetMetaData(depthMD);
            vector<Point> handContour;
            getHandContour(depthMat, skRHand, handContour);
            circle(depthMatBgr, Point(skRHand[0], skRHand[1]), 10, Scalar(skRHand[2],skRHand[2],skRHand[2]), CV_FILLED);
        }

		imshow("depthMatBgr", depthMatBgr);
		imshow("bgrMat", bgrMat);
		//imshow("undistortedImg",undistortedImg);
	}

    printf("Shutdown...");

    releaseSamplesWindow();
    context.Shutdown();

	return 0;
}
